/*
**	Manifests, Structures and Prototypes for the Client-Server
**	Chat Sample
*/
#include	<conio.h>
#include	<stdio.h>
#include	<errno.h>
#include	<grp.h>
#include	<pwd.h>
#include	<stdlib.h>
#include	<string.h>
#include	<unistd.h>

#include	<sys/dev.h>
#include	<sys/kernel.h>
#include	<sys/name.h>
#include	<sys/proxy.h>
#include	<sys/psinfo.h>
#include	<sys/sys_msg.h>
#include	<sys/types.h>


/*
**	Manifest Definitions
*/
#ifndef		TRUE
#define		TRUE		1
#define		FALSE		0
#endif

#define	CHATSERVER					"/qnx/ChatServer"

#define SERVER_VERSION_NUMBER		100
#define SERVER_VERSION_LETTER		'a'

#define	CLIENT_VERSION_NUMBER		100
#define CLIENT_VERSION_LETTER		'a'

#define MAX_CHAT_CLIENTS			16		/* Max Clients */
#define MAX_CHAT_MSGS				32		/* Msgs the Server should keep */
#define MAX_CLIENT_NAME_SIZE		32		/* Maximum Client Name */
#define MAX_CLIENT_MSG_SIZE			132		/* A single line of text */

#define CHATMSG						0x1000
#define RESERVED					0		/* RESERVED (_SYSMSG) */
#define	CHATMSG_SUBTYPE_OPEN		1		/* Open Request */
#define	CHATMSG_SUBTYPE_POST		2		/* Post a Message Request */
#define	CHATMSG_SUBTYPE_FETCH		3		/* Deliver New Message */
#define	CHATMSG_SUBTYPE_CLOSE		199		/* Close Request */

#define	MSG_KEYBOARD_READY			200		/* Keyboard Input (from Dev) */
#define MSG_MESSAGE_READY			210		/* New Message Available (from Server) */

#define OPEN_FLAG_SLOT_USED			0x0001	/* Client Open Table Slot Used */

#define	MSG_FLAG_ANOTHER_MSG		0x0001	/* Another Chat Message Available */

#define NEXT_MSG_SLOT( Cp, MBp )	((Cp)+1 == MBp + MAX_CHAT_MSGS ? MBp : (Cp) + 1)


/*
************************************************************
**		Chat Send/Receive Structures
************************************************************
*/
typedef struct							/* Standard Chat Send/Receive */
{	struct _sysmsg_hdr			Hdr;

}	CHAT_S_STANDARD_STRUCT;

typedef struct							/* Establish Chat Session */
{	struct _sysmsg_hdr			Hdr;
	pid_t						NewMsgProxy;
	unsigned short				zero1;		/* Possible future increase in pid size */
	char						MyName[ MAX_CLIENT_NAME_SIZE ];

}	CHAT_S_OPEN_STRUCT;

typedef struct							/* Keyboard/New Msg Event */
{	struct _sysmsg_hdr			Hdr;

}	CHAT_S_EVENT_STRUCT;

typedef struct							/* Post a Msg */
{	struct _sysmsg_hdr			Hdr;
	int							Handle;
	char						Msg[ MAX_CLIENT_MSG_SIZE + 1 ];
	
}	CHAT_S_POST_STRUCT;

typedef struct							/* Fetch a Msg */
{	struct _sysmsg_hdr			Hdr;
	int							Handle;

}	CHAT_S_FETCH_STRUCT;

typedef struct							/* Terminate Chat Session */
{	struct _sysmsg_hdr			Hdr;
	int							Handle;

}	CHAT_S_CLOSE_STRUCT;

/*
************************************************************
**		Chat Reply Structures
************************************************************
*/
typedef struct							/* Standard Chat Reply */
{	struct _sysmsg_hdr_reply		Hdr;

}	CHAT_R_STANDARD_STRUCT;

typedef struct							/* Establish Chat Session */
{	struct _sysmsg_hdr_reply		Hdr;
	int								Handle;

}	CHAT_R_OPEN_STRUCT;

typedef struct							/* Post a Msg */
{	struct _sysmsg_hdr_reply		Hdr;

}	CHAT_R_POST_STRUCT;

typedef struct							/* Fetch a Msg */
{	struct _sysmsg_hdr_reply		Hdr;
	unsigned						Flags;
	char							Msg[ MAX_CLIENT_NAME_SIZE + 2 + MAX_CLIENT_MSG_SIZE + 1 ];

}	CHAT_R_FETCH_STRUCT;

typedef struct							/* Terminate Chat Session */
{	struct _sysmsg_hdr_reply		Hdr;

}	CHAT_R_CLOSE_STRUCT;

typedef struct							/* Chat Error */
{	struct _sysmsg_hdr_reply		Hdr;

}	CHAT_R_ERROR_STRUCT;

typedef struct							/* Proc Death */
{	struct _sysmsg_hdr_reply		Hdr;

}	SYSMSG_R_DEATH_STRUCT;

typedef struct							/* sin Version */
{	struct _sysmsg_hdr_reply		Hdr;
	struct _sysmsg_version_reply	Body;

}	SYSMSG_R_VERSION_STRUCT;


/*
************************************************************
**		SEND/RECEIVE STRUCTURE
************************************************************
*/
typedef union
{	CHAT_S_STANDARD_STRUCT		Standard;
	CHAT_S_OPEN_STRUCT			Open;
	CHAT_S_EVENT_STRUCT			Event;
	CHAT_S_POST_STRUCT			Post;
	CHAT_S_FETCH_STRUCT			Fetch;
	CHAT_S_CLOSE_STRUCT			Close;

}	CHAT_S_MSG_STRUCT;


/*
************************************************************
**		REPLY STRUCTURE
************************************************************
*/
typedef union
{	CHAT_R_STANDARD_STRUCT		Standard;
	CHAT_R_OPEN_STRUCT			Open;
	CHAT_R_POST_STRUCT			Post;
	CHAT_R_FETCH_STRUCT			Fetch;
	CHAT_R_CLOSE_STRUCT			Close;
	CHAT_R_ERROR_STRUCT			Error;
	SYSMSG_R_DEATH_STRUCT		Death;
	SYSMSG_R_VERSION_STRUCT		Version;

}	CHAT_R_MSG_STRUCT;


/* End of File */
