#ifdef __USAGE

%C [-n chat_name]

Options:
  -n          specify the name to use during chat session.

Where:
  chat_name   is the chat name by which messages added by you will be
              tagged.  Default is login name.

#endif
/*
**	Sample Client
**
**	Implements a multi-user chat facility (like a party-line)
**	all participants see all other participants text.
**
**	This program implements the Client Portion.
**
*/
#include	"chat.h"


/*
******************************************************************************
**		LOCAL Data
******************************************************************************
*/
static char					*ProgramNamep;
static char					 ChatServer[] = CHATSERVER;


/*
******************************************************************************
**		LOCAL Helper Functions
******************************************************************************
*/
/****************************************************************************/
static nid_t		Pid2Nid( pid_t pid )
{	struct _psinfo	psdata;

	if ( (qnx_psinfo( PROC_PID, pid, &psdata, 0, NULL )) == -1 )
	{	fprintf( stderr, "%s: ERROR, process info, errno = \"%s\"\n",
							ProgramNamep, strerror( errno ) );
		exit( EXIT_FAILURE );
	}

	return psdata.un.vproc.remote_nid;
}

/****************************************************************************/
static void			RawOnDevMode( unsigned *old_mode )
{
	if ( (*old_mode = dev_mode( STDIN_FILENO, 0, _DEV_ECHO | _DEV_EDIT )) == -1 )
	{	fprintf( stderr, "%s: ERROR, Unable to turn raw on from dev mode, errno = \"%s\"\n", ProgramNamep, strerror( errno ) );
		exit( EXIT_FAILURE );
	}
}

/****************************************************************************/
static void			RawOffDevMode( unsigned *old_mode )
{
	if ( dev_mode( STDIN_FILENO, *old_mode, _DEV_ECHO | _DEV_EDIT ) == -1 )
	{	fprintf( stderr, "%s: ERROR, Unable to turn raw off from dev mode, errno = \"%s\"\n", ProgramNamep, strerror( errno ) );
		exit( EXIT_FAILURE );
	}
}

/****************************************************************************/
static void			ClearDevTrigger()
{
	if ( dev_state( STDIN_FILENO, 0, _DEV_EVENT_INPUT ) == -1 )
	{	fprintf( stderr, "%s: ERROR, Unable to clear dev's keyboard proxy, errno = \"%s\"\n", ProgramNamep, strerror( errno ) );
		exit( EXIT_FAILURE );
	}
}

/****************************************************************************/
static void			ArmDevTrigger( pid_t kb_proxy )
{
	if ( dev_arm( STDIN_FILENO, kb_proxy, _DEV_EVENT_INPUT ) == -1 )
	{	fprintf( stderr, "%s: ERROR, Unable to arm dev's keyboard proxy, errno = \"%s\"\n", ProgramNamep, strerror( errno ) );
		exit( EXIT_FAILURE );
	}
}


/*****************************************************************************
**	MAINLINE
*/
					main( int argc, char *argv[] )
{
	char				*name_to_use	= NULL;	/* Our chat name */
	int			 		 handle;				/* To be used in communicating with Server */
	pid_t		 		 kb_proxy,				/* Keyboard Ready Proxy */
				 		 msg_proxy,				/* New Message Ready Proxy */
						 msg_rem_proxy;			/* Remote New Message Proxy */
	pid_t		 		 server_pid;			/* Server Process ID */
	nid_t				 server_nid;			/* Server Node ID */
	int			 		 done;					/* Terminate flag */
	unsigned	 		 old_mode;				/* Save old console mode */
	CHAT_S_MSG_STRUCT	 s_msg;					/* Send message */
	CHAT_R_MSG_STRUCT	 r_msg;					/* Receive/Reply message */
	int					 i;

	extern char			*optarg;				/* Argument Processing */
	extern int			 opterr,
						 optopt;

	fprintf( stdout, "%s - v%c.%c%c%c (%s)\n",
			ProgramNamep = argv[ 0 ],
			'0' + CLIENT_VERSION_NUMBER/100,
			'0' + CLIENT_VERSION_NUMBER/10-CLIENT_VERSION_NUMBER/100*100/10,
			'0' + CLIENT_VERSION_NUMBER-CLIENT_VERSION_NUMBER/10*10,
			CLIENT_VERSION_LETTER,
			__DATE__ );

	opterr = FALSE;								/* handle our own arg errors */

	while( (i = getopt( argc, argv, "n:" )) != -1 )
		switch( i )
		{	case 'n':	name_to_use = optarg;	break;

			default:
				fprintf( stderr, "%s: Invalid Option '%c'\n", ProgramNamep, optopt ); 
				exit( 1 );
		}

	if ( ! name_to_use )
		name_to_use = getlogin();

	/*
	**	Find Chat Server
	*/
	if ( (server_pid = qnx_name_locate( (nid_t) 0, ChatServer, 0, NULL )) == -1 )
	{	fprintf( stderr, "%s: ERROR, can't find server \"%s\" (Server/nameloc Running?)\n", ProgramNamep, ChatServer );
		exit( EXIT_FAILURE );
	}

	/*
	**	Get Chat Server's Node
	*/
	server_nid = Pid2Nid( server_pid );

	/*
	**	Setup a Proxy for the Chat Server
	*/
	s_msg.Event.Hdr.type = MSG_MESSAGE_READY;
	if ( (msg_proxy = qnx_proxy_attach( 0, (char *) &s_msg.Event.Hdr.type, sizeof s_msg.Event.Hdr.type, -1 )) == -1 )
	{	fprintf( stderr, "%s: ERROR, Unable to attach local proxy, errno = \"%s\"\n", ProgramNamep, strerror( errno ) );
		exit( EXIT_FAILURE );
	}

	/*
	**	In case the Server is running on another node, create a
	**	remote proxy.
	*/
	if ( (msg_rem_proxy = qnx_proxy_rem_attach( server_nid, msg_proxy )) == -1 )
	{	fprintf( stderr, "%s: ERROR, Unable to attach remote proxy, errno = \"%s\"\n", ProgramNamep, strerror( errno ) );
		exit( EXIT_FAILURE );
	}

	/*
	**	Request a Chat Session:
	**		1. Tell server which proxy to use
	**		2. Give our "Chat" name
	*/
	s_msg.Open.Hdr.type		= CHATMSG;
	s_msg.Open.Hdr.subtype	= CHATMSG_SUBTYPE_OPEN;
	s_msg.Open.NewMsgProxy	= msg_rem_proxy;
	strcpy( s_msg.Open.MyName, name_to_use );
	if ( Send( server_pid, &s_msg.Open, &r_msg.Open, sizeof s_msg.Open, sizeof r_msg.Open ) == -1 )
	{	fprintf( stderr, "%s: ERROR, Unable to Send to Server, errno = \"%s\"\n", ProgramNamep, strerror( errno ) );
		exit( EXIT_FAILURE );
	}

	if ( r_msg.Open.Hdr.status != EOK )
	{	fprintf( stderr, "%s: ERROR, From Chat Server, errno = \"%s\"\n", ProgramNamep, strerror( r_msg.Open.Hdr.status ) );
		exit( EXIT_FAILURE );
	}

	handle = r_msg.Open.Handle;

	/*
	**	Setup a Proxy for Dev
	*/
	s_msg.Event.Hdr.type = MSG_KEYBOARD_READY;
	if ( (kb_proxy = qnx_proxy_attach( 0, (char *) &s_msg.Event.Hdr.type, sizeof s_msg.Event.Hdr.type, -1 )) == -1 )
	{	fprintf( stderr, "%s: ERROR, Unable to attach \"Keyboard Ready\" proxy, errno = \"%s\"\n", ProgramNamep, strerror( errno ) );
		exit( EXIT_FAILURE );
	}

	/*
 	**	Turn on raw input, clear then arm the trigger
	*/
	RawOnDevMode( &old_mode );
	ClearDevTrigger();
	ArmDevTrigger( kb_proxy );

	fprintf( stdout, "\tThe chat server is located on node %d\n", server_nid );
	fprintf( stdout, "\tYour chat name is \"%s\"\n", name_to_use );
	fprintf( stdout, "\tBegin typing to get message prompt, <cr> to post to \"%s\".\n", ChatServer );
	fprintf( stdout, "\t\".\"<cr> Terminates.\n\n" );
	fflush( stdout );

	done = FALSE;
	while ( ! done )
	{	Receive( 0, &s_msg, sizeof s_msg );

		switch ( s_msg.Event.Hdr.type )
		{	case MSG_KEYBOARD_READY:

            	RawOffDevMode( &old_mode );	/* Allow canonical processing */
				ClearDevTrigger();

				putch( '>' );
				putch( ' ' );
				gets( s_msg.Post.Msg );

				RawOnDevMode( &old_mode );
				ArmDevTrigger( kb_proxy );

				switch( *s_msg.Post.Msg )
				{	case '.':
						done = TRUE;

					case 0x0d:
						continue;
				}

				s_msg.Post.Hdr.type		= CHATMSG;
				s_msg.Post.Hdr.subtype	= CHATMSG_SUBTYPE_POST;
				s_msg.Post.Handle		= handle;
				if ( Send( server_pid, &s_msg.Post, &r_msg.Post, sizeof s_msg.Post, sizeof r_msg.Post ) == -1 )
				{	fprintf( stderr, "%s: ERROR, Sending to %s, errno = \"%s\"\n", ProgramNamep, ChatServer, strerror( errno ) );
					exit( EXIT_FAILURE );
				}

				if ( r_msg.Post.Hdr.status != EOK )
				{	fprintf( stderr, "%s: ERROR, Error from %s, errno = \"%s\"\n", ProgramNamep, ChatServer, strerror( r_msg.Post.Hdr.status ) );
					exit( EXIT_FAILURE );
				}
			break;

			case MSG_MESSAGE_READY:
				do
				{	s_msg.Fetch.Hdr.type	= CHATMSG;
					s_msg.Fetch.Hdr.subtype	= CHATMSG_SUBTYPE_FETCH;
					s_msg.Fetch.Handle		= handle;
					Send( server_pid, &s_msg.Fetch, &r_msg.Fetch, sizeof s_msg.Fetch, sizeof r_msg.Fetch );
					puts( r_msg.Fetch.Msg );
				}
				while ( r_msg.Fetch.Flags & MSG_FLAG_ANOTHER_MSG );
			break;

			default:
				fprintf( stderr, "%s: ERROR, Spurious Message!\n", ProgramNamep );
		}
	}
	s_msg.Close.Hdr.type	= CHATMSG;
	s_msg.Close.Hdr.subtype	= CHATMSG_SUBTYPE_CLOSE;
	s_msg.Close.Handle		= handle;
	Send( server_pid, &s_msg.Close, &r_msg.Close, sizeof s_msg.Close, sizeof r_msg.Close );

   	RawOffDevMode( &old_mode );	/* Allow canonical processing */
	qnx_proxy_rem_detach( server_nid, msg_rem_proxy );
	qnx_proxy_detach( msg_proxy );
	qnx_proxy_detach( kb_proxy );
}


/* End of File */
