#ifdef __USAGE

%C
  Program provides network wide chat facility.

#endif
/*
**	Sample Administrator
**
**		Implements a mult-user chat facility (like a party-line)
**	all participants see all other participants text.
**
**		This program implements the Server Portion.
**
*/
#include	"chat.h"


/*
******************************************************************************
**		LOCAL Structures
******************************************************************************
*/
typedef struct
{	unsigned					 MsgNo;			/* Incrementing Msg # */
	pid_t						 Pid;			/* Client Process ID */
	char						 Msg[ MAX_CLIENT_NAME_SIZE + 2 + MAX_CLIENT_MSG_SIZE + 1 ];

}	MSG_BOARD_STRUCT;

typedef struct
{	unsigned					 Flags;
	pid_t						 Pid;			/* Client Process ID */
	pid_t						 PPid;			/* Client's "New Msg" Proxy */
	char						 Name[ MAX_CLIENT_NAME_SIZE + 1 ];
	MSG_BOARD_STRUCT			*LastMsgp;		/* Last Msg Read Pointer */
	unsigned					 LastMsgNo;		/* Last Msg No (sanity check) */
	unsigned					 MsgsOutstanding;

}	OPEN_STRUCT;


/*
******************************************************************************
**		LOCAL Variables
******************************************************************************
*/
static char						*ProgramName;
static MSG_BOARD_STRUCT			 MsgBoardArray[ MAX_CHAT_MSGS ];
static OPEN_STRUCT				 ClientOpenArray[ MAX_CHAT_CLIENTS ];
static char						 ChatServer[] = CHATSERVER;


/*
******************************************************************************
**		LOCAL Helper Functions
******************************************************************************
*/
/****************************************************************************/
static int			AllocClientOpenSlot( OPEN_STRUCT coa[] )
{	OPEN_STRUCT				*cop,
							*co_lastp	= coa + MAX_CHAT_CLIENTS;

	for ( cop = coa ; cop < co_lastp ; cop++ )
		if ( (cop->Flags & OPEN_FLAG_SLOT_USED) == FALSE )
		{	memset( cop, 0, sizeof *cop );
	    	cop->Flags = OPEN_FLAG_SLOT_USED;
			break;
		}

	return cop < co_lastp ? cop - coa : -1;
}

/*****************************************************************************/
static OPEN_STRUCT	*DeAllocClientOpenSlot( OPEN_STRUCT coa[], pid_t pid )
{	OPEN_STRUCT				*cop,
							*co_lastp	= coa + MAX_CHAT_CLIENTS;

	for ( cop = coa ; cop < co_lastp ; cop++ )
		if ( cop->Pid == pid  &&  cop->Flags & OPEN_FLAG_SLOT_USED )
		{	cop->Flags &= ~OPEN_FLAG_SLOT_USED;
			break;
		}

	return cop < co_lastp ? cop : NULL;
}

/*****************************************************************************/
static int			ClientsAllGone( OPEN_STRUCT coa[] )
{	OPEN_STRUCT				*cop,
							*co_lastp	= coa + MAX_CHAT_CLIENTS;

	for ( cop = coa ; cop < co_lastp ; cop++ )
		if ( cop->Flags & OPEN_FLAG_SLOT_USED )
			return FALSE;

	return TRUE;
}

/*****************************************************************************/
static void		AddMsg( OPEN_STRUCT coa[], MSG_BOARD_STRUCT msg_boarda[], MSG_BOARD_STRUCT **hpp, MSG_BOARD_STRUCT **tpp, OPEN_STRUCT *clientp, char *mp )
{	OPEN_STRUCT				*cop,
							*co_lastp	= coa + MAX_CHAT_CLIENTS;
	MSG_BOARD_STRUCT		*hp			= *hpp,
							*tp			= *tpp;
	unsigned				 next_msg_no;

	/*
	**	Add the message to the board (hp should always point to the next
	**	available slot).
	*/
	hp->Pid = clientp ? clientp->Pid : NULL;
	sprintf( hp->Msg, "%u. %s: %s", hp->MsgNo, clientp ? clientp->Name : ChatServer, mp );

	/*
	**	Advance one slot in the Circular Message Board Buffer
	*/
	next_msg_no = hp->MsgNo + 1;

	*hpp = hp = NEXT_MSG_SLOT( hp, msg_boarda );	/* Update Head Pointer */

	hp->MsgNo = next_msg_no;

	if ( hp == tp )
		*tpp = tp = NEXT_MSG_SLOT( tp, msg_boarda );/* Update Tail Pointer */

	/*
	**	Possibly notify other clients
	*/
	for ( cop = coa ; cop < co_lastp ; cop++ )
		if ( cop->Flags & OPEN_FLAG_SLOT_USED  &&  (clientp->Pid != cop->Pid || clientp == NULL) )
			if ( cop->MsgsOutstanding++ == 0 )
				Trigger( cop->PPid );
}

/*****************************************************************************/
static void		GetNextMsg( MSG_BOARD_STRUCT msg_boarda[], OPEN_STRUCT *clientp, CHAT_R_MSG_STRUCT *r_msgp )
{	MSG_BOARD_STRUCT		*cmsgp = clientp->LastMsgp;

	if ( clientp->MsgsOutstanding )
	{	do
			cmsgp = NEXT_MSG_SLOT( cmsgp, msg_boarda );
		while ( cmsgp->Pid == clientp->Pid );		/* Ignore our own msgs */

		clientp->LastMsgp	= cmsgp;
		clientp->LastMsgNo	= cmsgp->MsgNo;			/* Sanity Check */
		clientp->MsgsOutstanding--;
		strcpy( r_msgp->Fetch.Msg, cmsgp->Msg );
	}
	else
		strcpy( r_msgp->Fetch.Msg, "!!! Msg Requested when none available !!!" );

	if ( clientp->MsgsOutstanding )
		r_msgp->Fetch.Flags	|=  MSG_FLAG_ANOTHER_MSG;
	else
		r_msgp->Fetch.Flags	&= ~MSG_FLAG_ANOTHER_MSG;
}

/*****************************************************************************/
static void		ReplyPidDescrepancy( CHAT_R_MSG_STRUCT *r_msg, pid_t pid )
{	r_msg->Error.Hdr.status = EPERM;
	Reply( pid, r_msg, sizeof *r_msg );
}


/*
******************************************************************************
**	MAINLINE
******************************************************************************
*/
				main( int argc, char *argv[] )
{	int						 NameID;		/* Name Server ID */
	pid_t					 pid;			/* Current Client's pid */
	OPEN_STRUCT				*clientp;		/* Current Client */
	MSG_BOARD_STRUCT		*headmbp,		/* Current Message Head */
							*tailmbp;		/* Current Message Tail */
	CHAT_S_MSG_STRUCT		 s_msg;			/* Message Sent from Client */
	CHAT_R_MSG_STRUCT		 r_msg;			/* Reply to Client */

	ProgramName = argv[0];

	if ( (NameID = qnx_name_attach( 0L, ChatServer )) == -1 )
	{	fprintf( stderr, "%s - ERROR:  Unable to register name (Server already running?), errno = \"%s\"\n", ProgramName, strerror( errno ) );
		exit( EXIT_FAILURE );
	}

	/*
	**	Apply for Server Status, Request Death Messages
	*/
	if ( qnx_pflags( ~0L, _PPF_SERVER | _PPF_INFORM, 0, 0 ) == -1 )
	{	fprintf( stderr, "%s - ERROR:  Unable to obtain server status, errno = \"%s\"\n", ProgramName, strerror( errno ) );
		exit( EXIT_FAILURE );
	}

	headmbp = tailmbp = MsgBoardArray;

	for( ;; )
	{	pid = Receive( 0, &s_msg, sizeof s_msg );

		switch ( s_msg.Standard.Hdr.type )
		{	case _SYSMSG:
				switch ( s_msg.Standard.Hdr.subtype )
				{	case _SYSMSG_SUBTYPE_DEATH:
						if ( clientp = DeAllocClientOpenSlot( ClientOpenArray, pid ) )
						{	if ( ClientsAllGone( ClientOpenArray ) )
								(headmbp = tailmbp = MsgBoardArray)->MsgNo = 0;
							else
							{	sprintf( s_msg.Post.Msg, "%s HAS UNEXPECTEDLY WITHDRAWN", clientp->Name );
								AddMsg( ClientOpenArray, MsgBoardArray, &headmbp, &tailmbp, NULL, s_msg.Post.Msg );
							}
						}
						r_msg.Death.Hdr.status = EOK;
						Reply( pid, &r_msg.Death, sizeof r_msg.Death );
					break;

					case _SYSMSG_SUBTYPE_VERSION:
						strcpy( r_msg.Version.Body.name, ChatServer );
						strcpy( r_msg.Version.Body.date, __DATE__ );
						r_msg.Version.Body.version	= SERVER_VERSION_NUMBER;
						r_msg.Version.Body.letter	= SERVER_VERSION_LETTER;
						r_msg.Version.Body.more		= FALSE;
						r_msg.Version.Hdr.status		= EOK;
						Reply( pid, &r_msg.Version, sizeof r_msg.Version );
					break;

					default:
						r_msg.Standard.Hdr.status	= ENOSYS;
						Reply( pid, &r_msg.Standard.Hdr.status, sizeof r_msg.Standard.Hdr.status );
				}
			break;

			case CHATMSG:
				switch ( s_msg.Standard.Hdr.subtype )
				{	case CHATMSG_SUBTYPE_OPEN:
						/*
						**	Four Step Process:
						**		1. Get an unused Client Open Entry from the Open Table
						**		2. Set up the Client Open Entry
						**		3. Post a message to the board on the client's behalf
						**		4. Reply with the handle to be used on subsequent requests
						*/
						if ( (r_msg.Open.Handle = AllocClientOpenSlot( ClientOpenArray )) == -1 )
							r_msg.Open.Hdr.status = ENOMEM;
						else
						{	clientp				= &ClientOpenArray[ r_msg.Open.Handle ];
							clientp->Pid		= pid;
							clientp->PPid		= s_msg.Open.NewMsgProxy;
							strncpy( clientp->Name, s_msg.Open.MyName, sizeof (clientp->Name) - 1 );
							clientp->Name[ sizeof (clientp->Name) - 1 ] = '\0';
							clientp->LastMsgp	= headmbp;
							clientp->LastMsgNo	= headmbp->MsgNo;
                		
							AddMsg( ClientOpenArray, MsgBoardArray, &headmbp, &tailmbp, clientp, "HAS JOINED SESSION" );

							r_msg.Open.Hdr.status	= EOK;
						}

						Reply( pid, &r_msg.Open, sizeof r_msg.Open );
					break;

					case CHATMSG_SUBTYPE_POST:
						if ( s_msg.Post.Handle > MAX_CHAT_CLIENTS || pid != (clientp = &ClientOpenArray[ s_msg.Post.Handle ])->Pid )
						{	ReplyPidDescrepancy( &r_msg, pid );
							continue;
						}
						AddMsg( ClientOpenArray, MsgBoardArray, &headmbp, &tailmbp, clientp, s_msg.Post.Msg );
						r_msg.Post.Hdr.status = EOK;
						Reply( pid, &r_msg.Post, sizeof r_msg.Post );
					break;

					case CHATMSG_SUBTYPE_FETCH:
						if ( s_msg.Post.Handle > MAX_CHAT_CLIENTS || pid != (clientp = &ClientOpenArray[ s_msg.Post.Handle ])->Pid )
						{	ReplyPidDescrepancy( &r_msg, pid );
							continue;
						}
						GetNextMsg( MsgBoardArray, clientp, &r_msg );
						r_msg.Fetch.Hdr.status = EOK;
						Reply( pid, &r_msg.Fetch, sizeof r_msg.Fetch );
					break;

					case CHATMSG_SUBTYPE_CLOSE:
						if ( s_msg.Post.Handle > MAX_CHAT_CLIENTS || pid != (clientp = &ClientOpenArray[ s_msg.Post.Handle ])->Pid )
						{	ReplyPidDescrepancy( &r_msg, pid );
							continue;
						}

						DeAllocClientOpenSlot( ClientOpenArray, pid );

						if ( ClientsAllGone( ClientOpenArray ) )
							(headmbp = tailmbp = MsgBoardArray)->MsgNo = 0;
						else
						{	sprintf( s_msg.Post.Msg, "%s HAS WITHDRAWN", clientp->Name );
							AddMsg( ClientOpenArray, MsgBoardArray, &headmbp, &tailmbp, NULL, s_msg.Post.Msg );
						}
						r_msg.Close.Hdr.status = EOK;
						Reply( pid, &r_msg.Close, sizeof r_msg.Close );
					break;

					default:
						r_msg.Standard.Hdr.status	= ENOSYS;
						Reply( pid, &r_msg.Standard.Hdr.status, sizeof r_msg.Standard.Hdr.status );
				}
			break;

			default:
				r_msg.Standard.Hdr.status	= ENOSYS;
				Reply( pid, &r_msg.Standard.Hdr.status, sizeof r_msg.Standard.Hdr.status );
		}
	}
	qnx_name_detach( 0, NameID );
}


/* End of File */
